#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "graph.h"
#include "utilities.h"

long n_vertices(FILE * map_file){
    long n;
    int i;
    i = 0;
    n = 0;
    while(i!= EOF && map_file){
        i = fscanf(map_file, "#Nodes: %ld\n", &n);
        if(i==1){
            return n;
        }
    }
    fprintf(stderr, "No node count found in map file.");
    fclose(map_file);
    exit(-1);
}

long n_edges(FILE * map_file){
    long n;
    int i;
    i = 0;
    n = 0;
    /*
    int c;
    while ((c = fgetc(map_file)) != EOF){
        putchar(c);
    }
    exit(-1);
    */
    while(i != EOF && map_file){
        i = fscanf(map_file, "#Edges: %ld\n", &n);
        if(i==1){
            return n;
        }
    }
    fprintf(stderr, "No edge count found in map file.");
    fclose(map_file);
    exit(-1);
}

static void initialize(Graph * g){
    g->n_vertices = 0;
    g->n_edges = 0;
    g->vertices = NULL;
    g->edges=NULL;
}

Graph * readGraph(char * map_path){
    long i;
    int junk;
    int x, c;
    int source, dest;
    char name[256];
    char * tmp;
    int ci;
    bool using_stdin;
    FILE * map_file;
    Graph * g = allocm(sizeof(*g));
    initialize(g);
    
    if(strcmp(map_path, "-")==0){
        map_file = stdin;
        using_stdin = TRUE;
    }
    else{
        map_file = fopen(map_path, "r");
        using_stdin = FALSE;
        if(!map_file){
            fprintf(stderr, "Error opening map file: %s", map_path);
            exit(-1);
        } 
    }
  

    /*Read in number of vertices*/
    g->n_vertices = n_vertices(map_file);
    g->vertices = allocm(g->n_vertices*sizeof(*g->vertices));

    for(i = 0; i < g->n_vertices; i++){
        /*Read line into vertex struct*/
        ci = 0;
        while(TRUE){
            name[ci] = getc(map_file);
            if(name[ci] == EOF){
                fprintf(stderr, "Error: input file truncated?");
                exit(-1);
            }
            if(name[ci] == '\n'){
                ci++;
                break;
            }
            ci++;
        }
        name[ci] = '\0';
        tmp = strrchr(name, '\t');
        tmp[0] = '\n';
        tmp[1] = '\0';
        g->vertices[i].n_successors = 0;
        g->vertices[i].successors = NULL;
        g->vertices[i].to_str = allocm((1 + strlen(name))*sizeof(char));
        strcpy(g->vertices[i].to_str, name);
        x = sscanf(name, "%d %lf %lf\n", &junk, &g->vertices[i].longitude, &g->vertices[i].latitude);
        /*Die if formatting incorrect*/
        if(x != 3){
            fprintf(stderr, "%d, %s\n", x, name);
            fprintf(stderr, "Fewer than expected nodes!\n");
            exit(-1);
        }

    }

    /*Read number of edges*/
    g->n_edges = n_edges(map_file);
    g->edges = allocm(g->n_edges*sizeof(*g->edges));

    for(i = 0; i < g->n_edges; i++){
        name[0] = '\0';
        x = fscanf(map_file, "%d %d %lf", &source, &dest, &g->edges[i].duration);

        if(x < 3){
            fprintf(stderr, "Fewer than expected edges! %d\n", x);
            exit(-1);
        }
        ci = 0;
        name[255] = '\0';
        while(ci < 255){
            name[ci] = getc(map_file);
            if(name[ci] == '\n'){
                name[ci] = '\0';
                break;
            }
            ci++; 
        }
        g->edges[i].source = g->vertices + source;
        g->edges[i].destination = g->vertices + dest;
        if (name[0] != '\0'){
            g->edges[i].name = allocm((strlen(name)+1)*sizeof(char));
            strcpy(g->edges[i].name, name);
        }
        else{
            g->edges[i].name = allocm(sizeof(char));
            *g->edges[i].name = '\0';
            while ((c = fgetc(map_file)) != '\n'){}
        }
        g->vertices[source].n_successors++;
    } 

    /*Allocate room for successor pointers*/
    for(i = 0; i < g->n_vertices; i++){
        g->vertices[i].successors = allocm(g->vertices[i].n_successors*sizeof(Edge *));
        g->vertices[i].n_successors = 0;
    }
    /*Fill in successor pointers*/
    for(i=0; i < g->n_edges; i++){
        g->edges[i].source->successors[g->edges[i].source->n_successors++] = g->edges + i;
    }
    if(!using_stdin){
        fclose(map_file);
    }
    return g;
}

void free_graph(Graph * g){
    long i;
    for(i = 0; i < g->n_edges;i++){
        free(g->edges[i].name);
    }
    free(g->edges);
    for(i =0; i < g->n_vertices; i++){
        free(g->vertices[i].successors);
        free(g->vertices[i].to_str);
    }
    free(g->vertices);
    free(g);
}

void emit_vertex(const Vertex * v, FILE * f){
    /*more verbose might be useful for debugging*/
    /*fprintf(f, "Lat:%f Lon:%f succ:%ld", v->latitude, v->longitude, v->n_successors);*/
    /*fprintf(f, "%f %f", v->latitude, v->longitude);*/
    fprintf(f, "%s", v->to_str);
}

void emit_edge(const Edge * e, FILE * f){
    fprintf(f, "s:");
    emit_vertex(e->source, f);
    fprintf(f, " -> ");
    emit_vertex(e->destination, f);
    fprintf(f, ", duration:%f %s", e->duration, e->name);
}

void emit_graph(const Graph * g, FILE * f){
    long i;
    fprintf(f, "%ld Nodes:\n", g->n_vertices);
    for(i=0;i<g->n_vertices;i++){
        emit_vertex(g->vertices + i, f);
        fprintf(f, "\n");
    }
    fprintf(f, "%ld Edges:\n", g->n_edges);
    for(i=0;i<g->n_edges;i++){
        emit_edge(g->edges + i, f);
        fprintf(f, "\n");
    }
}

void emit_metadata(const Metadata * m, FILE * f){
    fprintf(f, "%ld nodes expanded\n%ld nodes generated\n", m->nodes_expanded, m->nodes_generated);
}

Vertex * find_nearest(double lon, double lat, Graph * g){
    long i;
    double minimum_distance;  /* really distance squared */
    double distance;  /* really distance squared */
    Vertex * best = NULL;
    Vertex * cur;
    minimum_distance = HUGE_VAL;
    for(i=0; i<g->n_vertices; i++){
        cur = g->vertices + i;
        distance = dist2(cur->latitude - lat, cur->longitude - lon);
        if(best == NULL || distance < minimum_distance){
            best = cur;
            minimum_distance = distance;
        }
    }
    return best;
}