#pragma once
#include <stdio.h>

/* 
UNH CS 758 Assignment 10: Shortest Paths 
Devin Wild Thomas 11/2023

Sparse Graph datastructure for lat,lon graph
You should not need to modify this
*/

typedef struct Vertex Vertex;
typedef struct Edge Edge;

struct Vertex{
    double latitude;
    double longitude;
    long n_successors;
    Edge* * successors; /* Array of pointers to successor edges */
    char * to_str; /* String representation for printing, ignore */
};

struct Edge{ /* directed */
    Vertex * source;
    Vertex * destination;
    double duration;  /* In seconds */
    char * name; /* Street name from OSM, might be useful for debugging? */
};

typedef struct Graph{
    long n_vertices;
    long n_edges;
    Vertex * vertices; /* Array of vertices */
    Edge * edges;  /* Array of Edges */
} Graph;

typedef struct Metadata{ /* To measure performance increment when you generate/expand search nodes */
    long nodes_expanded;
    long nodes_generated;
} Metadata;

/* Printing functions, if you modify for debugging make sure you revert them to be compatible with the harness */
void emit_metadata(const Metadata * m, FILE * f);
void emit_vertex(const Vertex * v, FILE * f);
void emit_edge(const Edge * e, FILE * f);
void emit_graph(const Graph * g, FILE * f);

/* Reads the graph, should not need to be touched */
Graph * readGraph(char * map_path);

void free_graph(Graph * g);

/* Inefficient, only used to find the nearest match for the command line input source and goal. */
Vertex * find_nearest(double lat, double lon, Graph * g);